# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import run

from f1z1_common import is_validators

from .base import ArgsType, CoroOrFunction, IRunner, IRunnerStrategy, KwargsType
from .strategy import RunStrategies, StrategyFactory


class AsyncRunner(IRunner):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 args: ArgsType = None,
                 kwargs: KwargsType = None):
        self._coro_or_func = coro_or_func
        self._args = () if not args else args
        self._kwargs = {} if not kwargs else kwargs

    @property
    def coro_or_func(self):
        return self._coro_or_func

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    def run(self):
        return run(self.main())

    async def main(self):
        runner, kwargs = self, self.kwargs
        return await self._run_impl(runner, kwargs)

    async def _run_impl(self, runner: IRunner, kwargs: KwargsType):
        coro_or_func = runner.coro_or_func
        strategy: IRunnerStrategy = None
        if self._is_coro(coro_or_func):
            strategy = StrategyFactory.create(
                RunStrategies.CORO,
                runner
            )
        elif self._is_async_function(coro_or_func):
            strategy = StrategyFactory.create(
                RunStrategies.CORO_FUNCTION,
                runner,
                kwargs=kwargs
            )
        elif self._is_function(coro_or_func):
            strategy = StrategyFactory.create(
                RunStrategies.FUNCTION,
                runner
            )
        else:
            raise ValueError(
                f"coro_or_func need coroutine, async function or function, bu got {type(coro_or_func).__name__}"
            )
        return await strategy.async_run()

    def _is_coro(self, coro_or_func: CoroOrFunction):
        return is_validators.is_coroutine(coro_or_func)

    def _is_async_function(self, coro_or_func: CoroOrFunction):
        return is_validators.is_async_function(coro_or_func)

    def _is_function(self, coro_or_func: CoroOrFunction):
        return is_validators.is_function(coro_or_func)



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import ArgsType, CoroOrFunction, IRunner, IRunnerStrategy, KwargsType
from .arunner import AsyncRunner
from .api import set_loop_policy, is_runner, create_runner, start



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import set_event_loop_policy, AbstractEventLoopPolicy
from .base import ArgsType, CoroOrFunction, KwargsType, IRunner
from .arunner import AsyncRunner


def set_loop_policy(policy: AbstractEventLoopPolicy = None):
    if policy is None:
        return
    try:
        set_event_loop_policy(policy)
    except Exception as e:
        raise e


def is_runner(value):
    return isinstance(value, IRunner)


def create_runner(
        coro_or_func: CoroOrFunction,
        args: ArgsType = None,
        kwargs: KwargsType = None
):
    return AsyncRunner(
        coro_or_func,
        args=args,
        kwargs=kwargs
    )


def start(runner: IRunner):
    if not is_runner(runner):
        raise ValueError(f"runner need IRunner instance, but got {type(runner).__name__}")
    return runner.run()



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import get_event_loop, get_running_loop
from enum import Enum
from functools import partial

from f1z1_common import is_validators

from .base import KwargsType, IRunner, IRunnerStrategy


class CoroStrategy(IRunnerStrategy):
    """
    协程对象策略
    """

    def __init__(self, runner: IRunner):
        self._runner = runner

    @property
    def runner(self):
        return self._runner

    async def execute(self, runner: IRunner, **kwargs):
        return await runner.coro_or_func

    async def async_run(self, *args, **kwargs):
        return await self.execute(runner=self.runner)


class CoroFunctionStrategy(IRunnerStrategy):
    """
    协程函数策略
    """

    def __init__(self, runner: IRunner,
                 kwargs: KwargsType = None):
        self._runner = runner
        self._kwargs = kwargs if kwargs else {}

    @property
    def runner(self):
        return self._runner

    @property
    def kwargs(self):
        return self._kwargs

    def _check_coro_function(self, value) -> bool:
        pass

    async def execute(self, runner: IRunner, **kwargs):
        return await runner.coro_or_func(*runner.args, **kwargs)

    async def async_run(self, *args, **kwargs):
        return await self.execute(
            self.runner,
            **self.kwargs
        )


class FunctionStrategy(IRunnerStrategy):

    def __init__(self, runner: IRunner):
        self._runner = runner

    @property
    def runner(self):
        return self._runner

    async def execute(self, runner: IRunner, *args, **kwargs):
        loop = get_running_loop()
        if loop is None:
            loop = get_event_loop()
        return await  loop.run_in_executor(None, runner.coro_or_func, *runner.args)

    async def async_run(self, *args, **kwargs):
        return await self.execute(self.runner)


class RunStrategies(Enum):
    CORO = CoroStrategy
    CORO_FUNCTION = CoroFunctionStrategy
    FUNCTION = FunctionStrategy


class StrategyFactory(object):

    @classmethod
    def create(cls,
               strategies: RunStrategies,
               runner: IRunner,
               **kwargs) -> IRunnerStrategy:
        klass = cls._to_subclass(strategies)
        f = cls._factory(klass, runner)
        return f(**kwargs)

    @classmethod
    def _factory(cls, subclass: IRunnerStrategy, runner: IRunner):
        if not cls._is_strategy_subclass(subclass):
            raise ValueError(f"need a IRunnerStrategy subclass, but got {type(subclass).__name__}")
        return partial(subclass, runner)

    @staticmethod
    def _is_strategy_subclass(klass):
        return issubclass(klass, IRunnerStrategy)

    @staticmethod
    def _to_subclass(strategies: RunStrategies):
        if not is_validators.is_enum(strategies):
            raise ValueError(f"strategies need Enum instance, but got {type(strategies).__name__}")
        return strategies.value



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from typing import Any, Callable, Coroutine, Dict, Tuple, TypeVar, Union

ArgsType = Tuple[Any]
KwargsType = Dict[str, Any]
CoroOrFunction = Union[Callable, Coroutine]
ReturnType = TypeVar("ReturnType")


class IRunner(object):
    """
    runner interface
    """

    @property
    def coro_or_func(self) -> CoroOrFunction:
        raise NotImplementedError("coro_or_func")

    @property
    def args(self) -> ArgsType:
        raise NotImplementedError("")

    def run(self, *args, **kwargs) -> ReturnType:
        raise NotImplementedError("")


class IRunnerStrategy(object):
    """
    runner strategy
    """

    @property
    def runner(self) -> IRunner:
        raise NotImplementedError("")

    async def async_run(self, *args, **kwargs) -> ReturnType:
        raise NotImplementedError("")
